/*
	SPDX-License-Identifier: MIT
	Copyright © 2023-2025 Amebis
*/

#pragma once

#include "common.hpp"
#include <CoreFoundation/CoreFoundation.h>
#include <memory>
#include <system_error>

namespace macstd {
	///
	/// Core Foundation object traits
	///
	template <typename T>
	struct cfobject_traits
	{
		static inline constexpr T invalid = static_cast<T>(NULL);

		///
		/// Releases a Core Foundation object
		///
		/// \sa [CFRelease function](https://developer.apple.com/documentation/corefoundation/1521153-cfrelease)
		///
		static void free(T h)
		{
			CFRelease(h);
		}

	private:
		static T duplicate(T h);
	};

	///
	/// Core Foundation object traits
	///
	template <typename T>
	using cfobject = handle<T, cfobject_traits<T>>;

	using cfarray = cfobject<CFArrayRef>;
	using cfdictionary = cfobject<CFDictionaryRef>;
	using cferror = cfobject<CFErrorRef>;
	using cfstring = cfobject<CFStringRef>;

	///
	/// Core Foundation runtime error
	///
	class cf_runtime_error : public num_runtime_error<CFIndex>
	{
	protected:
		cferror m_error;

	public:
		///
		/// Constructs an exception
		///
		/// \param[in] error  Core Foundation error
		///
		cf_runtime_error(CFErrorRef error) :
			num_runtime_error<CFIndex>(CFErrorGetCode(error), message(error)),
			m_error(error)
		{
		}

		///
		/// Constructs an exception
		///
		/// \param[in] error  Core Foundation error
		///
		cf_runtime_error(cferror&& error) :
			num_runtime_error<CFIndex>(CFErrorGetCode(error), message(error)),
			m_error(std::move(error))
		{
		}

		CFErrorRef cferror() const noexcept { return m_error; }

	protected:
		///
		/// Returns a string explaining the meaning of a Core Foundation error
		///
		static std::string message(CFErrorRef error)
		{
			macstd::cfstring description(CFErrorCopyDescription(error));
			if (description.valid() && CFStringGetLength(description) > 0) {
				auto str = CFStringGetCStringPtr(description, kCFStringEncodingUTF8);
				if (str)
					return str;
			}
			std::string msg("Core Foundation error " + std::to_string(CFErrorGetCode(error)));
			macstd::cfstring domain(CFErrorGetDomain(error));
			if (domain.valid() && CFStringGetLength(domain) > 0) {
				auto str = CFStringGetCStringPtr(domain, kCFStringEncodingUTF8);
				if (str)
					msg += std::string(", domain ") + str;
			}
			return msg;
		}
	};
}
